-- dStruct.e
-- A library to (sort of) emulate structures within Euphoria

-- V1.0, created by Derek Parnell, February 2002.
--    version 1.1 July 2002
--      * Added destruct() routine.
--      * Added more documentation.

-- The basic idea is that we have two lists: one to store the structure definitions,
-- and another to store the structure instances.
-- 
-- The structure definitions are stored in a sequence called vKnownStructs. Each
-- item in this sequence represents either a definition of a structure, or a
-- deleted definition. Deleted definitions are integers (value of -1) and definitions
-- are sequences. 
-- A structure definition is a list of element definitions, and an element definition
-- is a sequence of three items - DataType, MaxSize, Name. 

-- DataType is one of the supported types - ATOM, INT, SEQ, OBJ, and STRUCTDefinition.
-- Only data values matching the element's type can be stored. In order words,
-- you cannot store a sequence in an ATOM element, nor an atom in a SEQ element.

-- MaxSize is only used for SEQ types and represents the maximum length of the
-- sequence. No instance of this structure can have a SEQ element lnger than this
-- maximum size. Note that a value of -1, means unlimited size.

-- Name is a standardized name given to the element. It's scope is the definition, 
-- which means that two different structure definitions can both have an element
-- with the same name.

-- Once created, a structure definition cannot be altered, it can be deleted though.
-- When deleting a structure definition, all instances of that structure are also
-- deleted. Referencing a deleted structure or any of its instances will cause a
-- typedef error and abort the program.
-- 
-- A structure is defined by one or more calls the element() routine, which adds
-- elements to the current structure being defined, and a call to struct() 
-- routine to complete the definition. The value returned by element() is used to
-- store and fetch data from instances of a structure, and the value returned by
-- struct() is used when creating instances of the structure.
--
-- Structure instances are stored in the sequence called vStructSpace. Each item
-- in this structure represents either an instances of a structure or a deleted
-- instance. Deleted instances are integers (value -1) and instances are sequences.
-- The instance sequence is a list of data values, one per element defined in the
-- structure, and one 'pointer' back to structure defininition. When an instances
-- is created, using the new() routine, its data values are initialized - INT, ATOM
-- and OBJ datatypes are set to zero, and SEQ datatypes are set to an empty sequence.
-- 
-- If an element of a structure is itself another structure, the new() routine 
-- automatically creates a new instance of the substructure and initializes the
-- parent structure's element to refer to the substructure instance. Substructures
-- are initialized recursively.
--
-- Thus is a structure definition was for an {INT, OBJ, SEQ}, the new instance
-- would be created and initialized to {0, 0, {}}.

-- You cannot create recursive structures. That is, while defining a structure, 
-- you cannot define one of its elements to be a structure of the type you are
-- in the process of defining. You can create a structure that allows its instances
-- to refer to other instances of the same structure type. This is done by using
-- the OBJ datatype, and store() and fetch() to manage its contents. 

-- Routines:                                      
--  element          : Defines an element of a structure
--  struct           : Completes the definition of a structure
--  new              : Creates a new instance of a structure
--  store            : Sets the value of one or more elements in a structure
--  fetch            : Gets the value of one or more elements in a structure
--  copy             : Copies the values of a structure to another structure
--  gettype          : Returns the type of object passed to it.
--  schema           : Returns a sequence representing the internal
--                      definition of the structure.
--  ElementNumber    : Returns the ordinal number for a structure element name
--  destroy          : Removes a instance of a structure
--  destruct         : Removes a structure definition and all instances based on it.
--  cStructToString  : Returns a string repesentation of a structure
--  cStructToSequence: Returns a sequence repesentation of a structure
--  cSequenceToString: Returns a string repesentation of a sequence
--  cNumberToString  : Returns a string repesentation of a number
--  initialize       : Reinitializes a structure's elements
--  QuoteString      : Returns a string enclosed in quotes with any
--                      embedded quotes escaped.
--  StandardName     : Standardizes a string that is meant to be an identifier.

-- type:
--    structure      :   Can be used for type-checking a structure variable
--    structdef      :   Can be used for type-checking a structdef variable

-- Constants:
--    ATOM          :   Identifies an element of type 'atom'
--    INT           :   Identifies an element of type 'integer'
--    SEQ           :   Identifies an element of type 'sequence'
--    OBJ           :   Identifies a generic object of any type.
--    STRUCTDefinition: Identifies a structure definition
--    STRUCTInstance: Identifies a structure instance.
--without trace

include wildcard.e
global constant
    ATOM = 1,
    INT  = 2,
    SEQ  = 3,
    OBJ  = 4,
    STRUCTInstance = "sI",
    STRUCTDefinition = "sD"
        
constant
    vElementTypes = {ATOM, INT, SEQ, OBJ, STRUCTDefinition}

-- Delimiters used to seperate elements during conversion to strings.
sequence vElementDelim vElementDelim = ","
integer vStringDelim vStringDelim = '"'
integer vEscapeLead vEscapeLead = #5C -- Backslash char.

-- List of struct instances
sequence vStructSpace vStructSpace = {}

-- List of struct definitions
sequence vKnownStructs vKnownStructs = {}

-- Pointer to structure currently being defined.
integer vCurrentStruct vCurrentStruct = 0

----------------------------------------------------------------------------
global function gettype(object pObject)
--   Returns a sequence containing the type of the supplied object.
--     Format of return is {DataType (,Option)}
--      where DataType is either INT, ATOM, SEQ, OBJ, or STRUCTDefinition
--      when STRUCTDefinition the return has a second element. If
--      this is zero then the object is a structdef object, otherwise
--      it is a structure object.
----------------------------------------------------------------------------
    object lSD
    sequence lRC

    lRC = {}
    -- First test for a structure instance.
    if sequence(pObject) and
       length(pObject) = 2 and
       equal(pObject[1], STRUCTInstance) and
       integer(pObject[2]) and
       pObject[2] <= length(vStructSpace) and
       pObject[2] > 0 and
       sequence(vStructSpace[pObject[2]]) then

        lSD = vStructSpace[pObject[2]][length(vStructSpace[pObject[2]])]
        if integer(lSD) and
           lSD <= length(vKnownStructs) and 
           lSD > 0 and
           sequence(vKnownStructs[lSD]) and
           length(vStructSpace[pObject[2]]) = length(vKnownStructs[lSD])+1 then

           return  {STRUCTDefinition,lSD}
        end if
    end if

    -- Next, test for a structure definition.
    if sequence(pObject) and
        length(pObject) = 2 and
        equal(pObject[1], STRUCTDefinition) and
        integer(pObject[2]) and
        pObject[2] <= length(vKnownStructs) and
        pObject[2] > 0 and
        sequence(vKnownStructs[pObject[2]]) then

           return {STRUCTDefinition, 0}
    end if

    -- Next, test for a native euphoria types.
    if sequence(pObject) then
        return {SEQ}
    elsif integer(pObject) then
        return {INT}
    elsif atom(pObject) then
        return {ATOM}
    end if

    return {OBJ}
end function

----------------------------------------------------------------------------
global function QuoteString(sequence pStr)
--   Returns the string, enclosed in quotes and embedded quotes escaped.
----------------------------------------------------------------------------
    integer lPos
    integer lMark

    -- Is there any quotes in the current string?
    lPos = find(vStringDelim, pStr)
    if lPos != 0 then
        -- Yes!
        -- Define a position marker for this string
        lMark = -1000000
        for j = -1 to -999999 by -1 do
            if find(j, pStr) = 0 then
                lMark = j
                exit
            end if
        end for

        -- Replace all quotes with position marker
        while lPos != 0 do
            pStr[lPos] = lMark
            lPos = find(vStringDelim, pStr)
        end while

        -- Replace all position markers with escaped-quote
        lPos = find(lMark, pStr)
        while lPos != 0 do
            pStr = pStr[1..lPos-1] & vEscapeLead & vStringDelim & pStr[lPos+1..length(pStr)]
            lPos = find(lMark, pStr)
        end while

    end if

    return vStringDelim & pStr & vStringDelim
end function

----------------------------------------------------------------------------
global type structure(object s)
--   Type Checking function for structure
----------------------------------------------------------------------------
    sequence lRC

    lRC = gettype(s)
    if (length(lRC) = 2) and 
       equal(lRC[1],STRUCTDefinition) and
       (lRC[2] != 0) then
        return 1
    else
        return 0
    end if
end type

----------------------------------------------------------------------------
global type structdef(object s)
--   Type Checking function for structure
----------------------------------------------------------------------------
    sequence lRC

    lRC = gettype(s)
    if (length(lRC) = 2) and 
       equal(lRC[1],STRUCTDefinition) and
       (lRC[2] = 0) then
        return 1
    else
        return 0
    end if
end type

----------------------------------------------------------------------------
global function StandardName(sequence pName)
--   Creates a standardize version of an identifying name.
--
--   Removes all whitespace
--   Converts to lowercase
----------------------------------------------------------------------------
    integer lT
    integer lS
    integer lE
    integer lI

    lS = 1
    lE = length(pName)
    lT = 1
    lI = 0
    while lT <= lE do
        if not find(pName[lT], " \t\n\r") then
            lI += 1
            pName[lI] = pName[lT]
        end if
        lT += 1
    end while

    return lower(pName[1 .. lI])

end function

----------------------------------------------------------------------------
global function element(object pType)
--   Defines an element of a structure
-- Returns integer: The ordinal number of the element within its structure.
--   If there is any problems, it returns zero.
----------------------------------------------------------------------------

    if integer(pType) then
        if find(pType, vElementTypes) then
            pType = {pType, -1, ""}
        else
            return 0
        end if
    elsif atom(pType) then
        return 0
    else
        if length(pType) > 3 then
            return 0
        end if
        -- Special case of {structdef,name}
        if length(pType) = 2 and
           sequence(pType[1]) and
           sequence(pType[2]) and
           length(pType[1])= 2 and
           equal(pType[1][1], STRUCTDefinition) then

            pType = {pType[1][1], pType[1][2], pType[2]}

        end if

        if not find(pType[1], vElementTypes) then
            return 0
        end if

        if length(pType) = 1 then
            pType = {pType[1], -1, ""}
        
        elsif length(pType) = 2 then
            if sequence(pType[2]) then
                pType = {pType[1], -1, pType[2]}
        
            elsif integer(pType[2]) then
                pType = {pType[1], pType[2], ""}
        
            else
                return 0
            end if
        
        else -- length is 3
            if sequence(pType[2]) and integer(pType[3]) then
                pType = {pType[1], pType[3], pType[2]}
        
            elsif integer(pType[2]) and sequence(pType[3]) then
        
                -- this is good.
            else
                return 0
            end if
        end if
    end if

    -- ASSERT length(pType) = 3,
    --        find(pType[1], vElementTypes) = true,
    --        integer(pType[2]),
    --        sequence(pType[3])

    -- Add on the standardized version of the element name.
    pType = append(pType, StandardName(pType[3]))
                
    -- If this is the first element of a new structure, then
    -- create the new empty structure definition first.
    if vCurrentStruct = 0 then
        -- First check for any deleted definitions.
        vCurrentStruct = find(-1, vKnownStructs)
        if vCurrentStruct = 0 then                 
            -- None, so need to create a new one.
            vKnownStructs = append(vKnownStructs, {})
            vCurrentStruct = length(vKnownStructs)
        else                                     
            -- Reuse the space that was deleted.
            vKnownStructs[vCurrentStruct] = {}            
        end if
    end if

    vKnownStructs[vCurrentStruct] = append(vKnownStructs[vCurrentStruct],
                                            pType)

    return length(vKnownStructs[vCurrentStruct])
end function

----------------------------------------------------------------------------
global function struct()
--   Completes the definition of a structure
-- Returns a structdef object, or {} if no structure is currently being
-- built
----------------------------------------------------------------------------
    integer lNewStruct
    
    if vCurrentStruct > 0 then
        lNewStruct = vCurrentStruct
        vCurrentStruct = 0
        return {STRUCTDefinition, lNewStruct }
    else
        return {}
    end if
end function

----------------------------------------------------------------------------
global procedure initialize(structure pS)
--   Reinitializes a structure's elements
----------------------------------------------------------------------------
    integer lSD
    object lDT

    lSD = vStructSpace[pS[2]][length(vStructSpace[pS[2]])]

    for i = 1 to length(vKnownStructs[lSD]) do
        lDT = vKnownStructs[lSD][i][1]
        if find(lDT,{ATOM, INT, OBJ}) then
            vStructSpace[pS[2]][i] = 0

        elsif equal(lDT,SEQ) then
            vStructSpace[pS[2]][i] = {}

        elsif equal(lDT, STRUCTDefinition) then
            initialize(vStructSpace[pS[2]][i])

        end if
    end for

end procedure

----------------------------------------------------------------------------
global function new(structdef pS)
--   Creates a new instance of a structure
----------------------------------------------------------------------------
    integer lIndex
    sequence lStruct

    -- Locate an empty slot
    lIndex = find(-1, vStructSpace)
    -- If we can't find an empty one, add a new slot
    if lIndex = 0 then
        vStructSpace = append(vStructSpace, 0)
        lIndex = length(vStructSpace)
    end if

    -- Create an 'empty' structure
    vStructSpace[lIndex] = repeat(0,1+length(vKnownStructs[pS[2]]))
    -- Insert the back pointer to the structure definition.
    vStructSpace[lIndex][length(vStructSpace[lIndex])] = pS[2]
    -- Build the return data
    lStruct = {STRUCTInstance, lIndex}

    -- Check for sub structures defined in this structure,
    -- and create instances of them too.
    for i = 1 to length(vKnownStructs[pS[2]]) do
        if equal(vKnownStructs[pS[2]][i][1], STRUCTDefinition) then
            vStructSpace[lIndex][i] = new(vKnownStructs[pS[2]][i][1..2])
        end if
    end for
    initialize(lStruct)

    return lStruct
end function

----------------------------------------------------------------------------
global function cNumberToString(atom pNumber)
--   Converts a number to a string representation of that number.
----------------------------------------------------------------------------
    sequence lTemp
    integer lS, lE

    if integer(pNumber) then
        return sprintf("%d", pNumber)
    end if
    -- I'm doing it this way to avid scientific notation, leading zeros,
    -- and trailing zeros.
    lTemp = sprintf("%15.15f", pNumber)
    lS = 1
    while lS <= length(lTemp) and lTemp[lS] = '0' do
        lS += 1
    end while
    -- Avoid a leading dot.
    if lTemp[lS] = '.' then
        lS -= 1
    end if

    lE = length(lTemp)
    while lE > lS and lTemp[lE] = '0' do
        lE -= 1
    end while
    -- Avoid a trailing dot.
    if lTemp[lE] = '.' then
        lE -= 1
    end if
    return lTemp[lS..lE]
end function

----------------------------------------------------------------------------
global function cSequenceToString(sequence pS)
--   Returns a string repesentation of a sequence.
----------------------------------------------------------------------------
    sequence lResults
    integer lS
    integer lE
    integer lT
    sequence lNewStr

    -- Test for character strings (converting non-printables as we go).
    lResults = ""
    lNewStr = pS
    lE = length(pS)
    lS = 1
    lT = 1
    for i = 1 to length(pS) do
        if not integer(pS[i]) or
           pS[i] < 0 or
           pS[i] > 255 then
            exit
        end if
        if pS[i] < ' ' then
            lNewStr = lNewStr[1..lT-1] & '^' & pS[i] + '@' & lNewStr[lT+1 .. lE]
            lE += 1
            lT += 1
        elsif pS[i] > '~' then
            lNewStr = lNewStr[1..lT-1] & '^' & sprintf("%03d",pS[i]) & lNewStr[lT+1 .. lE]
            lE += 3
            lT += 3
        end if
        lT += 1
    end for

    if lT > length(pS) then
        return lNewStr
    end if

    for i = 1 to length(pS) do
        lResults &= vElementDelim
        if integer(pS[i]) then
            lResults &= cNumberToString(pS[i])
        elsif atom(pS[i]) then
            lResults &= cNumberToString(pS[i])
        else
            lResults &= cSequenceToString(pS[i])
        end if
    end for

    return lResults[length(vElementDelim)+1 .. length(lResults)]
end function

----------------------------------------------------------------------------
global function cStructToString(structure pS)
--   Returns a string repesentation of a structure
----------------------------------------------------------------------------
    sequence lResults
    integer lNumElems
    integer lSD
    object lDT
    sequence lNV

    lSD = vStructSpace[pS[2]][length(vStructSpace[pS[2]])]
    lNumElems = length(vKnownStructs[lSD])

    lResults = ""

    -- Get each requested element's data.
    for i = 1 to lNumElems do
        lResults &= vElementDelim
        lDT = vKnownStructs[lSD][i][1]

        if equal(lDT, STRUCTDefinition) then
            lNV = cStructToString(vStructSpace[pS[2]][i])

        elsif find(lDT,{ATOM, INT}) then
            lNV = cNumberToString(vStructSpace[pS[2]][i])

        elsif equal(lDT,SEQ) then
            lNV = cSequenceToString(vStructSpace[pS[2]][i])
            lNV = QuoteString(lNV)
        else
            if atom(vStructSpace[pS[2]][i]) then
                lNV = cNumberToString(vStructSpace[pS[2]][i])
            else
                lNV = cSequenceToString(vStructSpace[pS[2]][i])
                lNV = QuoteString(lNV)
            end if
        end if

        lResults &= lNV
    end for

    -- Return everything except the leading delimiter
    return lResults[length(vElementDelim)+1 .. length(lResults)]
end function

----------------------------------------------------------------------------
global function cStructToSequence(structure pS)
--   Returns a sequence repesentation of a structure
----------------------------------------------------------------------------
    sequence lResults
    integer lNumElems
    integer lSD

    lSD = vStructSpace[pS[2]][length(vStructSpace[pS[2]])]
    lNumElems = length(vKnownStructs[lSD])

    -- Preallocate space
    lResults = repeat(0, lNumElems)

    -- Get each requested element's data.
    for i = 1 to lNumElems do
        if equal(vKnownStructs[lSD][i][1],STRUCTDefinition) then
            lResults[i] = cStructToSequence(vStructSpace[pS[2]][i])
        else
            lResults[i] = vStructSpace[pS[2]][i]
        end if
    end for

    return lResults
end function

----------------------------------------------------------------------------
global function ElementNumber(object pS, object pElement)
--   Returns the ordinal number for a structure element name.
--           If the name is not identifiable, zero is returned.
--  Parameter 1 can be either an integer (ordinal struct def),
--              a structdef object or a structure object.
----------------------------------------------------------------------------
    integer lSD
    object lType

    -- Determine what the param 1 is.
    lSD = 0
    lType = gettype(pS)
    if lType[1] = INT then
        lSD = pS

    elsif equal(lType[1], STRUCTDefinition) then
        if lType[2] = 0 then
            lSD = pS[2]
        else
            lSD = vStructSpace[pS[2]][length(vStructSpace[pS[2]])]
        end if
    end if

    if lSD = 0 then
        return 0
    end if

    if sequence(pElement) and length(pElement) > 0 and
       not equal(pElement[1], STRUCTDefinition) then
       -- scan for the element name in its definition.
       pElement = StandardName(pElement)
       for j = 1 to length(vKnownStructs[lSD]) do
            if equal(pElement, vKnownStructs[lSD][j][4]) then
                -- Replace name reference with ordinal reference
                return j
            end if
       end for
    end if
    return 0
end function

----------------------------------------------------------------------------
global procedure store(structure pS, sequence pElement)
--   Sets the value of one or more elements in a structure
----------------------------------------------------------------------------
    integer lSD
    object lDT
    integer lSE
    integer lSI
    integer lMaxSize
    object  lNV

    -- Get the Structure Instance pointer
    lSI = pS[2]
    -- Get the Structure Definition pointer
    lSD = vStructSpace[lSI][length(vStructSpace[lSI])]

    -- Cater for the simple case of only one element being stored.
    if length(pElement) = 2 then
        if integer(pElement[1]) then
            lSE = pElement[1]
        else
            lSE = ElementNumber(lSD, pElement[1])
        end if
        if lSE != 0 then
            pElement = {{lSE, pElement[2]}}
        end if
    end if

    -- Process each element change request
    for i = 1 to length(pElement) do
        -- A change request must be a 2-element sequence,
        --  in the form {ElementID, NewValue}
        if sequence(pElement[i]) and length(pElement[i]) = 2 then
            -- Get the Structure Element id
            if integer(pElement[i][1]) then
                lSE = pElement[i][1]
            else
                lSE = ElementNumber(lSD, pElement[i][1])
            end if
            -- Only use valid ones.
            if lSE > 0 and lSE <= length(vKnownStructs[lSD]) then

                -- Get the element's DataType
                lDT = vKnownStructs[lSD][lSE][1]
                -- Get the New Value
                lNV = pElement[i][2]

                -- Only do the assign if the actual datatype matches the
                -- one in the definition.
                if find(lDT,{ATOM,OBJ}) and atom(lNV) then
                    vStructSpace[lSI][lSE] = lNV

                elsif find(lDT,{INT,OBJ}) and integer(lNV) then
                    vStructSpace[lSI][lSE] = lNV

                elsif find(lDT,{SEQ,OBJ}) and sequence(lNV) then
                    -- Check for size-restricted sequences.
                    lMaxSize = vKnownStructs[lSD][lSE][2]
                    if lMaxSize != -1 and length(lNV) > lMaxSize then
                        lNV = lNV[1..lMaxSize]
                    end if
                    vStructSpace[lSI][lSE] = lNV

                elsif equal(lDT, STRUCTDefinition ) then
                    -- Recursively store substructures.
                    store(vStructSpace[lSI][lSE], lNV)
                else -- assume an OBJ
                    vStructSpace[lSI][lSE] = lNV
                end if
            end if
        end if
    end for

end procedure

----------------------------------------------------------------------------
global function fetch(structure pS, object pElement)
--   Gets the value of one or more elements in a structure
----------------------------------------------------------------------------
    sequence lResults
    integer lSD
    integer lSI
    integer lSE

    -- Get the Structure Instance pointer
    lSI = pS[2]
    -- Get the Structure Definition pointer
    lSD = vStructSpace[lSI][length(vStructSpace[lSI])]

    -- Cater for the simple case of one element being fetched.
    if integer(pElement) then
        pElement = {pElement}
    else
        lSE = ElementNumber(lSD, pElement)
        if lSE > 0 then
            pElement = {lSE}
        elsif sequence(pElement) and length(pElement) = 2 and
               sequence(pElement[2]) and
               (integer(pElement[1]) or ElementNumber(lSD,pElement[1])) then
               pElement = {pElement}
        end if
    end if

    -- Preallocate space
    lResults = repeat(0, length(pElement))

    -- Get each requested element's data.
    for i = 1 to length(pElement) do

        -- See if the element is being referenced by name.
        if integer(pElement[i]) then
            lSE = pElement[i]
        else
            lSE = ElementNumber(lSD, pElement[i])
        end if

        if lSE > 0 and lSE <= length(vKnownStructs[lSD]) then
            if equal(vKnownStructs[lSD][lSE][1], STRUCTDefinition) then
                lResults[i] = cStructToString(vStructSpace[lSI][lSE])
            else
                lResults[i] = vStructSpace[lSI][lSE]
            end if
        -- Test for a substructure reference
        elsif sequence(pElement[i]) and
              length(pElement[i]) = 2 and
              equal(pElement[i][1],STRUCTDefinition) and
              integer(pElement[i][2]) then
                lResults[i] = fetch(vStructSpace[lSI][pElement[i][1]],
                                     pElement[i][2])
        -- Test for a subelement reference.
        elsif sequence(pElement[i]) and
              length(pElement[i]) = 2 and
              sequence(pElement[i][2]) then

            if integer(pElement[i][1]) then
                lSE = pElement[i][1]
            else
                lSE = ElementNumber(lSD, pElement[i][1])
            end if
            if lSE > 0 then
                lResults[i] = fetch(vStructSpace[lSI][lSE],
                                 pElement[i][2])
            else
                lResults[i] = 0
            end if
        else
            lResults[i] = 0
        end if
    end for

    if length(pElement) = 1 then
        return lResults[1]
    else
        return lResults
    end if
end function

----------------------------------------------------------------------------
global procedure copy(structure pSource, structure pTarget)
--   Copies the values of a structure to another structure
----------------------------------------------------------------------------
    integer lSD
    integer lx

    -- Don't need to do anything to copy to oneself.
    if equal(pSource, pTarget) then
        return
    end if

    -- Can only copy to a the structure type.
    lSD = vStructSpace[pSource[2]][length(vStructSpace[pSource[2]])]
    lx  = vStructSpace[pTarget[2]][length(vStructSpace[pTarget[2]])]
    if lx != lSD then
        return
    end if

    for i = 1 to length(vKnownStructs[lSD]) do
        if equal(vKnownStructs[lSD][i][1], STRUCTDefinition) then
            vStructSpace[pTarget[2]][i] = new(vKnownStructs[lSD][i][1..2])
            copy(vStructSpace[pSource[2]][i],vStructSpace[pTarget[2]][i] )
        else
            vStructSpace[pTarget[2]][i] = vStructSpace[pSource[2]][i]
        end if
    end for
end procedure


----------------------------------------------------------------------------
global procedure destroy(structure pS)
--   Removes a instance of a structure
----------------------------------------------------------------------------
    integer lSD
    object lDT
    integer lSI

    -- Get the Structure Instance pointer
    lSI = pS[2]
    -- Get the Structure Definition pointer
    lSD = vStructSpace[lSI][length(vStructSpace[lSI])]

    -- Process each element change request
    for i = 1 to length(vKnownStructs[lSD]) do
        -- Get the element's DataType
        lDT = vKnownStructs[lSD][i][1]

        -- Recursively destroy substructures.
        if equal(lDT,STRUCTDefinition) then
            destroy(vStructSpace[lSI][i])
        end if
    end for

    vStructSpace[lSI] = -1
end procedure

----------------------------------------------------------------------------
global procedure destruct(structdef pSD)
--   Deletes the definition of a structure
----------------------------------------------------------------------------

    integer lSD
    
    lSD = pSD[2]
            
    -- Destroy any instances of this structure.               
    for i = 1 to length(vStructSpace) do
        if sequence(vStructSpace[i]) and
           vStructSpace[i][length(vStructSpace[i])] = lSD then
            destroy ( {STRUCTInstance, i} )
        end if
    end for
    
    vKnownStructs[lSD] = -1
    
end procedure

----------------------------------------------------------------------------
global function schema(object pS)
--   Returns a sequence containing the internal schema definition of the structure
----------------------------------------------------------------------------
    integer lSD
    sequence lType

    lType = gettype(pS)
    if length(lType) = 2 and equal(lType[1], STRUCTDefinition) then
        if lType[2] = 0 then
            lSD = pS[2]
        else
            lSD = vStructSpace[pS[2]][length(vStructSpace[pS[2]])]
        end if

        return vKnownStructs[lSD]
    else
        return {}
    end if
end function

